import numpy as np
import csv
from itertools import compress

from aurt.file_system import safe_open
from aurt.calibration_aux import find_nonstatic_start_and_end_indices

plot_colors = ['red', 'green', 'blue', 'chocolate', 'crimson', 'fuchsia', 'indigo', 'orange']

JOINT_N = "#"


class RobotData:
    """
    This class contains sampled robot data related to an experiment.
    """
    def __init__(self, file_path, delimiter, desired_timeframe=None, interpolate_missing_samples=False):
        self.n_joints = None  # TODO: automatically determine number of joints from csv data
        self.fields = [
            f"timestamp",
            f"target_q_{JOINT_N}",
            f"target_qd_{JOINT_N}",
            f"target_qdd_{JOINT_N}",
            f"target_current_{JOINT_N}",
            f"actual_current_{JOINT_N}",
            f"target_moment_{JOINT_N}",
            f"actual_q_{JOINT_N}",
            f"actual_qd_{JOINT_N}",
        ]
        self._load_data(file_path,
                         desired_timeframe=desired_timeframe,
                         interpolate_missing_samples=interpolate_missing_samples,
                         delimiter=delimiter)
        qd_target = np.array([self.data[f"target_qd_{j}"] for j in range(1, self.n_joints + 1)])
        self.non_static_start_index, self.non_static_end_index = find_nonstatic_start_and_end_indices(qd_target)

    def _load_data(self,
                    file_path,
                    desired_timeframe=None,
                    interpolate_missing_samples=False,
                    delimiter=' '):
        """
        Loads the robot data.
        """

        self._load_raw_csv_data(file_path=file_path, delimiter=delimiter)
        self.time -= self.time[0]  # Normalizing time range
        self._ensure_data_consistent()

        # Trim data to the desired timeframe
        if desired_timeframe is not None:
            self._trim_data(desired_timeframe)
            self._ensure_data_consistent()

        self._process_missing_samples(interpolate_missing_samples)
        self._ensure_data_consistent()

    def _process_missing_samples(self, interpolate_missing_samples):
        """
        Corrects lost samples by interpolating data at lost samples
        """

        # If the dataset has missing samples, find;
        #   1. the time instants / sample indices having missing samples and
        #   2. the period of time missing / number of missing samples
        # and interpolate (linearly) the data for all missing samples.

        self.dt_actual = self.time[1:] - self.time[:-1]
        assert len(self.dt_actual) == (len(self.time) - 1)
        (values, counts) = np.unique(self.dt_actual, return_counts=True)
        self.dt_nominal = values[np.argmax(counts)]  # Most frequently occurring dt
        sample_is_missing = np.array([abs(number - self.dt_nominal) > 1E-6 for number in self.dt_actual])
        time_missing = self.dt_actual - self.dt_nominal
        samples_missing = [round(float(element / self.dt_nominal)) for element in time_missing]  # list of ints

        len_data_before_interpolation = len(self.time)

        assert len(samples_missing) == (len_data_before_interpolation-1)

        n_samples_to_add = sum(samples_missing)

        if any(sample_is_missing):
            print(f"Warning: {n_samples_to_add}/{len_data_before_interpolation+n_samples_to_add} ({round(n_samples_to_add/(len_data_before_interpolation+n_samples_to_add) * 100, 2)} %) samples are missing in this dataset.")

        if interpolate_missing_samples:
            assert "interpolated" not in self.data
            self.data["interpolated"] = np.zeros(len_data_before_interpolation, dtype=bool)

            # Inserts values in data at indices of missing samples
            def insert_val(lst, idx_insert_before, v, nsamples):
                """
                Inserts a list of NaNs in data[f] at index 'idx_insert_after' having length 'samples_missing[k]'
                """
                return np.insert(lst, idx_insert_before, [v for _ in range(nsamples)])

            for k in reversed(range(len(sample_is_missing))):
                if sample_is_missing[k]:  # if a sample is missing at index k
                    idx_insert_before = k + 1
                    # Expand data and insert NaN's at locations for missing samples
                    for f in self.data:
                        if f != "interpolated":
                            self.data[f] = insert_val(self.data[f], idx_insert_before, float("nan"), samples_missing[k])
                    # Flag new insertions as interpolated
                    self.data["interpolated"] = insert_val(self.data["interpolated"], idx_insert_before, True, samples_missing[k])
                    # Also add timestamps as nan
                    self.time = insert_val(self.time, idx_insert_before, float("nan"), samples_missing[k])

            # The following two assertions ensure that the right number of samples were added.
            assert len(self.time) == len_data_before_interpolation + n_samples_to_add
            self._ensure_data_consistent()

            def nan_helper(y):
                """Helper to handle indices and logical indices of NaNs.

                Input:
                    - y, 1d numpy array with possible NaNs
                Output:
                    - nans, logical indices of NaNs
                    - index, a function, with signature indices = index(logical_indices),
                      to convert logical indices of NaNs to 'equivalent' indices
                Example:
                    # linear interpolation of NaNs
                    nans, x = nan_helper(y)
                    y[nans] = np.interp(x(nans), x(~nans), y[~nans])
                """

                return np.isnan(y), lambda z: z.nonzero()[0]

            # Interpolate NaN values
            nans, idx = nan_helper(self.time)

            def interpolate_nans(col):
                for index, value in zip(idx(nans), np.interp(idx(nans), idx(~nans), list(compress(col, ~nans)))):
                    col[index] = value

            for f in self.data:
                if f != "interpolated":
                    interpolate_nans(self.data[f])

            interpolate_nans(self.time)

    def _load_raw_csv_data(self, file_path, delimiter):
        with safe_open(file_path, mode='r') as csvFile:
            csv_reader = csv.DictReader(csvFile, delimiter=delimiter)

            self.n_joints = len([i for i in csv_reader.fieldnames if "target_q_" in i])

            self.time = np.empty(0)
            self.data = {}
            for f in self.fields:
                if JOINT_N in f:
                    for j in range(1, self.n_joints + 1):
                        f_j = f.replace(JOINT_N, str(j))
                        self.data[f_j] = np.empty(0)
                else:
                    self.data[f] = np.empty(0)

            for row in csv_reader:
                if not row: continue  # eliminates empty lines
                assert "timestamp" in row, f"Expected timestamp to be in the data. Perhaps you got the wrong delimiter? This is the row {row}"
                self.time = np.append(self.time, float(row["timestamp"]))
                for f in self.fields:
                    if JOINT_N in f:
                        for j in range(1, self.n_joints + 1):
                            f_j_out = f.replace(JOINT_N, str(j))
                            index_in = j - 1
                            f_j_in = f.replace(JOINT_N, str(index_in))
                            assert f_j_in in row, f"Problem indexing row with field {f_j_in}. Existing fields are: {row.keys()}."
                            self.data[f_j_out] = np.append(self.data[f_j_out], float(row[f_j_in]))
                    else:
                        self.data[f] = np.append(self.data[f], float(row[f]))

    def _trim_data(self, desired_timeframe):
        start_idx = 0
        while self.time[start_idx] < desired_timeframe[0]:
            start_idx += 1

        end_idx = len(self.time) - 1
        while self.time[end_idx] > desired_timeframe[1]:
            end_idx -= 1

        assert start_idx <= end_idx

        self.time = self.time[start_idx:end_idx + 1]

        for f in self.data.keys():  # crop all fields in data
            self.data[f] = self.data[f][start_idx:end_idx + 1]
        
    def _ensure_data_consistent(self):
        for f in self.data.keys():
            assert len(self.data[f]) == len(self.time), f"Field {f} in data is inconsistent. Expected {len(self.time)} samples. Got {len(self.data[f])} instead."

    def plot_missing_samples(self):
        import matplotlib.pyplot as plt
        plt.plot(self.data["interpolated"])
        plt.show()

    def plot_target_trajectory(self):
        q_m = np.array([self.data[f"actual_q_{j}"] for j in range(1, self.n_joints + 1)])
        qd_m = np.gradient(q_m, self.dt_nominal, edge_order=2, axis=1)
        qdd_m = (q_m[:, 2:] - 2 * q_m[:, 1:-1] + q_m[:, :-2]) / (self.dt_nominal ** 2)  # two fewer indices than q and qd

        # q,qd,qdd = central_finite_difference_and_crop(order=2, idx_start, idx_end)

        t = self.data["timestamp"]
        qd_m = qd_m[:, 1:-2]

        if idx_end == -1:
            qdd_end_idx = qd_tf.shape[1]
        else:
            qdd_end_idx = idx_end-1

        qdd_tf = qdd_tf[:, idx_start - 1:qdd_end_idx]
        qdd_m = qdd_m[:, -1:-1], np.append(qdd_m, None, axis=1)

        import matplotlib.pyplot as plt

        _, axs = plt.subplots(3, 1, sharex='all')
        axs[0].set(ylabel='Position [rad]')
        axs[1].set(ylabel='Velocity [rad/s]')
        axs[2].set(ylabel='Acceleration [rad/s^2]')

        for j in range(self.n_joints):
            # Actual
            axs[0].plot(t, q_m[j, :], ':', color=plot_colors[j], label=f"actual_{j}")
            axs[1].plot(t, qd_m[j, :], ':', color=plot_colors[j], label=f"actual_{j}")
            axs[2].plot(t, qdd_m[j, :], ':', color=plot_colors[j], label=f"actual_{j}")
            # Target
            axs[0].plot(t, self.data[f"target_q_{j+1}"], color=plot_colors[j], label=f"target_{j}")
            axs[1].plot(t, self.data[f"target_qd_{j+1}"], color=plot_colors[j], label=f"target_{j}")
            axs[2].plot(t, self.data[f"target_qdd_{j+1}"], color=plot_colors[j], label=f"target_{j}")

        for ax in axs:
            ax.legend()


        for j in range(self.n_joints):
            plt.plot(self.data["timestamp"], self.data[f"target_q_{j+1}"], color=plot_colors[j])

        plt.show()



import argparse
from logging.config import fileConfig
import logging
import numpy as np
import os.path
import aurt.api as api
from aurt.file_system import from_cache

def setup_logger(args):
    if args.logger_config is not None:
        fileConfig(args.logger_config)
    else:
        logging.basicConfig(level=logging.INFO)
    return logging.getLogger("aurt")


def compile_rbd(args):
    l = setup_logger(args)
    l.info("Compiling rigid body dynamics model.")
    l.debug(f"Gravity vector: {args.gravity}")

    if args.mdh[-4:] != ".csv":
        raise Exception(f"The provided mdh file {args.mdh} is not a CSV file. AURT only supports CSV files. Please provide a CSV file.")
    if not os.path.isfile(args.mdh):
        raise OSError(f"The mdh file {args.mdh} could not be located. Please specify a valid filename (csv).")
    gravity_vector_type = {True if isinstance(g,float) or isinstance(g,int) else False for g in args.gravity}
    if False in gravity_vector_type:
        raise TypeError(f"The given gravity vector is not a float, nor an integer. Please provide a valid gravity vector.")
    filename = from_cache(args.out + ".pickle")
    if os.path.isfile(filename):
        l.warning(f"The rigid body dynamics file {filename} already exists, and its content will be overwritten.")
    
    mdh_path = args.mdh
    gravity = np.array(args.gravity)
    output_path = args.out
    plotting = args.plot

    api.compile_rbd(mdh_path, gravity, output_path, plotting)


def compile_rd(args):
    l = setup_logger(args)
    l.info("Compiling robot dynamics model.")
    l.debug(f"Viscous friction powers: {args.friction_viscous_powers}.")

    filename = from_cache(args.model_rbd + ".pickle")
    if not os.path.isfile(filename):
        raise Exception(f"The rigid body dynamics file {filename} could not be located. Please specify a valid filename.")

    filename = from_cache(args.out + ".pickle")
    if os.path.isfile(filename):
        l.warning(f"The robot dynamics filename {args.out} already exists, and its content will be overwritten.")

    model_rbd_path = args.model_rbd
    friction_load_model = args.friction_load_model
    friction_viscous_powers = args.friction_viscous_powers
    #friction_hysteresis_model = args.friction_hysteresis_model # saved for later implementation
    output_path = args.out

    api.compile_rd(model_rbd_path, friction_load_model, friction_viscous_powers, output_path)


def calibrate(args):
    l = setup_logger(args)
    l.info("Calibrating robot dynamics model.")

    filename = from_cache(args.model + ".pickle")
    if not os.path.isfile(filename):
        raise Exception(f"The robot dynamics file {filename} could not be located. Please specify a valid filename.")

    if not os.path.isfile(args.data):
        raise Exception(f"The data file {args.data} could not be located. Please specify a valid filename.")

    if os.path.isfile(args.out_params):
        l.warning(f"The parameters filename {args.out_params} already exists, and its content will be overwritten.")

    filename = from_cache(args.out_calibration_model + ".pickle")
    if os.path.isfile(filename):
        l.warning(f"The calibration model filename {args.out_calibration_model} already exists, and its content will be overwritten.")

    model_path = args.model
    data_path = args.data
    params_path = args.out_params
    calbration_model_path = args.out_calibration_model
    plotting = args.plot
    
    api.calibrate(model_path, data_path, params_path, calbration_model_path, plotting)


def predict(args):
    l = setup_logger(args)
    l.info("Predicting robot current.")

    filename = from_cache(args.model + ".pickle")
    if not os.path.isfile(filename):
        raise Exception(f"The robot calibration file {filename} could not be located. Please specify a valid filename.")
    
    if not os.path.isfile(args.data):
        raise Exception(f"The data file {args.data} could not be located. Please specify a valid filename.")

    filename = from_cache(args.prediction)
    if os.path.isfile(filename):
        l.warning(f"The output prediction file {filename} already exists, and its content will be overwritten.")

    model_path = args.model
    data_path = args.data
    output_path = args.prediction
    
    api.predict(model_path, data_path, output_path)


def calibrate_validate(args):
    l = setup_logger(args)
    l.info("Calibrating and validating robot dynamics model.")

    filename = from_cache(args.model + ".pickle")
    if not os.path.isfile(filename):
        raise Exception(f"The robot dynamics file {filename} could not be located. Please specify a valid filename.")
    
    if not os.path.isfile(args.data):
        raise Exception(f"The data file {args.data} could not be located. Please specify a valid filename.")

    if os.path.isfile(args.out_params):
        l.warning(f"The parameters filename {args.out_params} already exists, and its content will be overwritten.")

    filename = from_cache(args.output_prediction)
    if os.path.isfile(filename):
        l.warning(f"The output prediction file {filename} already exists, and its content will be overwritten.")

    filename = from_cache(args.out_calibration_model + ".pickle")
    if os.path.isfile(filename):
        l.warning(f"The calibration model filename {args.out_calibration_model} already exists, and its content will be overwritten.")


    if not (0.1 < args.calibration_data_rel < 0.9):
        raise Exception(f"The calibration data rel value is not within the limits of 0.1 and 0.9, it is {args.calibration_data_rel}. Please provide a valid value.")

    model_path = args.model
    data_path = args.data
    calbration_model_path = args.out_calibration_model
    calibration_data_rel = args.calibration_data_rel # TODO: make sure to check that it is between 0 and 1
    plotting = args.plot
    params_path = args.out_params
    output_prediction_path = args.output_prediction
    
    api.calibrate_validate(model_path, data_path, calibration_data_rel, params_path, calbration_model_path, output_prediction_path, plotting)


def create_cmd_parser():
    # Command parser
    subparsers, args_parser = _init_cmd_parser()

    # sub commands
    ## compile_rbd
    _create_compile_rbd_parser(subparsers)
    ## compile_rd
    _create_compile_rd_parser(subparsers)
    ## calibrate
    _create_calibrate_parser(subparsers)
    ## predict
    _create_predict_parser(subparsers)
    ## validate
    _create_calibrate_validate_parser(subparsers)

    # Force help display when error occurrs. See https://stackoverflow.com/questions/3636967/python-argparse-how-can-i-display-help-automatically-on-error
    args_parser.usage = args_parser.format_help().replace("usage: ", "")

    args = args_parser.parse_args()

    args.command(args)

def _init_cmd_parser():
    args_parser = argparse.ArgumentParser(add_help=True)

    args_parser.add_argument('--logger-config', type=open,
                             help="Logger configuration file.")
    # args_parser.add_argument('command', choices=['compile-rbd', 'compile-rd', 'calibrate', 'predict'])
    # args_parser.add_argument('-h', help="Show help.", action="store_true")

    subparsers = args_parser.add_subparsers(
        help="Command to execute. Type 'aurt CMD --help' for more help about command CMD.")
    return subparsers, args_parser

def _create_compile_rbd_parser(subparsers):
    compile_rbd_parser = subparsers.add_parser("compile-rbd")

    compile_rbd_parser.add_argument('--mdh', required=True,
                                    help="Modified Denavit Hartenberg (MDH) parameters file (csv)")

    compile_rbd_parser.add_argument('--gravity', required=True,
                                    nargs=3,
                                    type=float,
                                    metavar='R',
                                    help="Gravity vector. Ex: 0 0 -9.81")

    compile_rbd_parser.add_argument('--out', required=True,
                                    help="Path of outputted rigid body dynamics model (pickle).")

    compile_rbd_parser.add_argument('--plot', action="store_true", default=False)

    compile_rbd_parser.set_defaults(command=compile_rbd)
    return compile_rbd_parser

def _create_compile_rd_parser(subparsers):
    compile_rd_parser = subparsers.add_parser("compile-rd")

    compile_rd_parser.add_argument('--model-rbd', required=True,
                                       help="The rigid body dynamics model created with the compile-rbd command.")

    compile_rd_parser.add_argument('--friction-load-model', required=True,
                                       choices=["none", "square", "absolute"],
                                       help="The friction load model.")

    compile_rd_parser.add_argument('--friction-viscous-powers', required=True, nargs="+",
                                       type=float,
                                       metavar='R',
                                       help="The viscous friction polynomial powers.")

    # compile_rd_parser.add_argument('--friction-hysteresis-model', required=True,
    #                                    choices=["sign", "maxwells"],
    #                                    help="The friction hysteresis model.")

    compile_rd_parser.add_argument('--out', required=True,
                                       help="Path of outputted robot dynamics model (pickle).")

    compile_rd_parser.set_defaults(command=compile_rd)
    return compile_rd_parser

def _create_calibrate_parser(subparsers):
    calibrate_parser = subparsers.add_parser("calibrate")

    calibrate_parser.add_argument('--model', required=True,
                                    help="The robot dynamics model created with the compile-rd command.")

    calibrate_parser.add_argument('--data', required=True,
                                    help="The measured data (csv).")

    calibrate_parser.add_argument('--out-params',
                                    help="The resulting parameter values (csv).")

    calibrate_parser.add_argument('--out-calibration-model',
                                    help="Path of the outputted robot calibration model (pickle).")

    calibrate_parser.add_argument('--plot', action="store_true", default=False)

    calibrate_parser.set_defaults(command=calibrate)
    return calibrate_parser

def _create_predict_parser(subparsers):
    predict_parser = subparsers.add_parser("predict")

    predict_parser.add_argument('--model', required=True,
                                    help="The calibration model created with the calibrate command.")

    predict_parser.add_argument('--data', required=True,
                                    help="The measured data (csv).")

    predict_parser.add_argument('--prediction', required=True,
                                    help="Path of outputted prediction values (csv).")

    predict_parser.set_defaults(command=predict)
    return predict_parser

def _create_calibrate_validate_parser(subparsers):
    calibrate_validate_parser = subparsers.add_parser("calibrate-validate")

    calibrate_validate_parser.add_argument('--model', required=True,
                                           help="The robot dynamics model created with the compile-rd command.")

    calibrate_validate_parser.add_argument('--data', required=True, help="The measured data (csv).")

    calibrate_validate_parser.add_argument('--calibration-data-rel', required=True, type=float,
                                           help="The relative fraction of the dataset to be used for calibration. The value should be in the range [0.1; 0.9].")

    calibrate_validate_parser.add_argument('--out-params', help="The resulting parameter values (csv).")

    calibrate_validate_parser.add_argument('--out-calibration-model',
                                           help="Path of the outputted robot calibration model (pickle).")

    calibrate_validate_parser.add_argument('--output-prediction', required=True,
                                           help="Path of outputted prediction values (csv).")

    calibrate_validate_parser.add_argument('--plot', action="store_true", default=False)

    calibrate_validate_parser.set_defaults(command=calibrate_validate)
    return calibrate_validate_parser

def main():
   create_cmd_parser()


if __name__ == '__main__':
    main()



import sys
import sympy as sp
import numpy as np


class JointDynamics:
    def __init__(self, n_joints, load_model=None, hysteresis_model=None, viscous_powers=None):
        self.n_joints = n_joints
        self.__qd = [sp.Integer(0)] + [sp.symbols(f"qd{j}") for j in range(1, self.n_joints + 1)]
        self.__tauJ = sp.symbols([f"tauJ{j}" for j in range(self.n_joints + 1)])

        # Defaults
        if load_model is None:
            load_model = 'square'  # 'none', 'square' or 'abs'
        if hysteresis_model is None:
            hysteresis_model = 'sign'  # 'sign' or 'gms'
        if viscous_powers is None:
            viscous_powers = [1]  # [1, 2, 3, ...], i.e. list of positive integers

        self.load_model = load_model
        self.hysteresis_model = hysteresis_model
        self.viscous_friction_powers = viscous_powers

    def __coulomb_friction_parameters(self):
        """
        Returns a list of 'n_joints + 1' elements, each element comprising a list of Coulomb friction parameters for
        each joint.
        """

        assert self.load_model.lower() in ('none', 'square', 'abs')

        fc = sp.symbols([f"Fc{j}" for j in range(self.n_joints + 1)])
        fcl = sp.symbols([f"Fcl{j}" for j in range(self.n_joints + 1)])

        if self.load_model.lower() == 'none':
            return [[fc[j]] for j in range(self.n_joints + 1)]
        else:
            return [[fc[j], fcl[j]] for j in range(self.n_joints + 1)]

    def __viscous_friction_parameters(self):
        """
        Returns a list of 'Njoints + 1' elements, each element comprising a list of viscous friction parameters for each
        joint.
        """

        return [[sp.symbols(f"Fv{j}_{i}") for i in self.viscous_friction_powers] for j in range(self.n_joints + 1)]

    def parameters(self):
        """
        Returns a list of 'Njoints + 1' elements with each element comprising a list of joint dynamics parameters for that
        corresponding joint.
        """

        fcs = self.__coulomb_friction_parameters()
        fvs = self.__viscous_friction_parameters()
        return [[*fcs[j], *fvs[j]] for j in range(1, self.n_joints + 1)]

    def number_of_parameters(self):
        par = self.parameters()
        return [len(par[j]) for j in range(self.n_joints)]

    def __coulomb_friction_basis(self):
        """
        Returns the basis (regressor) of the Coulomb friction model.
        """

        assert self.load_model.lower() in ('none', 'square', 'abs')

        h = self.__generalized_hysteresis_model()
        if self.load_model.lower() == 'none':
            return [[h[j]] for j in range(self.n_joints + 1)]
        elif self.load_model.lower() == 'square':
            return [[h[j], h[j] * self.__tauJ[j] ** 2] for j in range(self.n_joints + 1)]
        else:  # load_model == 'abs'
            return [[h[j], h[j] * abs(self.__tauJ[j])] for j in range(self.n_joints + 1)]

    def __generalized_hysteresis_model(self):
        """
        Returns the hysteresis model used in the Coulomb friction model.
        """
        assert self.hysteresis_model.lower() in ('sign', 'gms')

        if self.hysteresis_model.lower() == 'sign':
            return [sp.sign(self.__qd[j]) for j in range(self.n_joints + 1)]
        else:
            print(f"GMS model not yet implemented - using a simple static model instead.")
            return [sp.sign(self.__qd[j]) for j in range(self.n_joints + 1)]

    def __viscous_friction_basis(self):
        """
        Returns the basis (regressor) of the viscous friction model.
        """

        fv_basis = [[sp.Integer(0)] * len(self.viscous_friction_powers) for _ in range(self.n_joints + 1)]
        for j in range(self.n_joints + 1):
            for i, power in enumerate(self.viscous_friction_powers):
                if power % 2 == 0:  # even
                    fv_basis[j][i] = self.__qd[j] ** (power - 1) * abs(self.__qd[j])
                else:  # odd
                    fv_basis[j][i] = self.__qd[j] ** power
        return fv_basis

    def observation_matrix_joint(self, j, qd_j_num, tauJ_j_num):
        """Returns a (n_samples x self.number_of_parameters[j]) observation matrix"""
        assert qd_j_num.size == tauJ_j_num.size
        assert 0 <= j < self.n_joints

        n_samples = qd_j_num.size
        args_sym = [self.__qd[j+1], self.__tauJ[j+1]]
        args_num = np.concatenate((qd_j_num[:, np.newaxis], tauJ_j_num[:, np.newaxis]), axis=1).transpose()
        regressor_j = self.regressor()[j]
        observation_matrix_j = np.zeros((n_samples, regressor_j.shape[1]))
        sys.setrecursionlimit(int(1e6))
        reg_j_fcn = sp.lambdify(args_sym, regressor_j, 'numpy')
        observation_matrix_j[:n_samples, :] = reg_j_fcn(*args_num).squeeze().transpose()

        return observation_matrix_j

    def observation_matrix(self, qd_num, tauJ_num):
        """Returns a list of observation matrices for each joint."""
        assert qd_num.shape == tauJ_num.shape

        return [self.observation_matrix_joint(j, qd_num[j, :], tauJ_num[j, :]) for j in range(self.n_joints)]

    def regressor(self):
        """
        Returns an 'n_joints' list of regressors.
        """

        fc_basis = self.__coulomb_friction_basis()
        fv_basis = self.__viscous_friction_basis()
        return [sp.Matrix([*fc_basis[j], *fv_basis[j]]).T for j in range(1, self.n_joints + 1)]

    def dynamics(self):
        """
        Returns a list of joint dynamics with the list elements corresponding to the joint dynamics of each joint.
        """
        jd_basis = self.regressor()
        jd_par = self.parameters()
        return [sum([b * p for b, p in zip(jd_basis[i], jd_par[i])]) for i in range(len(jd_basis))]



from aurt.robot_dynamics import RobotDynamics
from aurt.signal_processing import central_finite_difference
import numpy as np
from scipy import signal
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
import pickle

from aurt.file_system import cache_numpy, cache_csv, load_numpy, from_cache, cache_object, load_object
from aurt.robot_data import RobotData, plot_colors


class RobotCalibration:
    def __init__(self, rd_filename, robot_data_path, relative_separation_of_calibration_and_prediction=None,
                 robot_data_predict=None):

        # Load saved RobotDynamics model
        filename = from_cache(rd_filename + ".pickle")
        with open(filename, 'rb') as f:
            self.robot_dynamics: RobotDynamics = pickle.load(f)


        if relative_separation_of_calibration_and_prediction is None and robot_data_predict is None:
            self.robot_data_calibration = RobotData(robot_data_path, delimiter=' ', interpolate_missing_samples=True)
            self.robot_data_validation = None
        elif relative_separation_of_calibration_and_prediction is not None and robot_data_predict is None:
            assert 0 < relative_separation_of_calibration_and_prediction < 1, "The specified relative separation of " \
                                                                              "data used for calibration and " \
                                                                              "prediction must be in the range from 0 " \
                                                                              "to 1."
            dummy_data = RobotData(robot_data_path, delimiter=' ')
            t_sep = dummy_data.time[-1] * relative_separation_of_calibration_and_prediction
            self.robot_data_calibration = RobotData(robot_data_path, delimiter=' ', interpolate_missing_samples=True, desired_timeframe=(0, t_sep))
            self.robot_data_validation = RobotData(robot_data_path, delimiter=' ', interpolate_missing_samples=True, desired_timeframe=(t_sep, np.inf))
            # self.robot_data_calibration._trim_data(desired_timeframe=(0, t_sep))
            # self.robot_data_validation._trim_data(desired_timeframe=(t_sep, np.inf))
        elif relative_separation_of_calibration_and_prediction is None and robot_data_predict is not None:
            self.robot_data_calibration = RobotData(robot_data_path, delimiter=' ', interpolate_missing_samples=True)
            self.robot_data_validation = robot_data_predict
        else:
            print("A wrong combination of arguments was provided.")

        self.f_dyn = 10  # Approx. cut-off frequency [Hz] of robot dynamics to be estimated
        self.downsampling_factor = round(0.8 / (4 * self.f_dyn * self.robot_data_calibration.dt_nominal))
        self.parameters = None
        self.estimated_output = None
        self.number_of_samples_in_downsampled_data = None

    def _measurement_vector(self, robot_data, start_index=1, end_index=-1):
        def compute_measurement_vector():
            i = np.array([robot_data.data[f"actual_current_{j}"] for j in range(1, self.robot_dynamics.n_joints + 1)]).T
            i_pf = RobotCalibration._parallel_filter(i, robot_data.dt_nominal, self.f_dyn)[start_index:end_index, :]
            i_pf_ds = RobotCalibration._downsample(i_pf, self.downsampling_factor)
            return i_pf_ds.flatten(order='F')  # y = [y1, ..., yi, ..., yN],  yi = [yi_{1}, ..., yi_{n_samples}]

        return compute_measurement_vector()

    def _observation_matrix(self, robot_data, start_index=1, end_index=-1):
        q_m = np.array([robot_data.data[f"actual_q_{j}"] for j in range(1, self.robot_dynamics.n_joints + 1)])

        # Low-pass filter (smoothen) measured angular position and obtain 1st and 2nd order time-derivatives
        q_tf, qd_tf, qdd_tf = RobotCalibration._trajectory_filtering_and_central_difference(q_m,
                                                                                             robot_data.dt_nominal,
                                                                                             self.f_dyn,
                                                                                             start_index,
                                                                                             end_index)

        n_samples_ds = self._measurement_vector(robot_data, start_index=start_index, end_index=end_index).shape[
                           0] // self.robot_dynamics.n_joints  # No. of samples in downsampled data
        observation_matrix = np.zeros((self.robot_dynamics.n_joints * n_samples_ds,
                                       sum(self.robot_dynamics.number_of_parameters())))  # Initialization
        for j in range(self.robot_dynamics.n_joints):
            # Obtain the rows of the observation matrix related to joint j
            obs_mat_j = self.robot_dynamics.observation_matrix_joint(j, q_tf, qd_tf, qdd_tf)

            # Parallel filter and decimate/downsample the rows of the observation matrix related to joint j.
            obs_mat_j_ds = RobotCalibration._downsample(
                RobotCalibration._parallel_filter(obs_mat_j, robot_data.dt_nominal, self.f_dyn),
                self.downsampling_factor)
            observation_matrix[j * n_samples_ds:(j + 1) * n_samples_ds, :] = obs_mat_j_ds
        return observation_matrix

    def calibrate(self, filename_parameters):  # , calibration_method='wls', weighting='variance'):
        # TODO: make it possible to specify the relative portion of the dataset you want, e.g. 0.5 for half of the
        #  dataset. Also, the actual number of e.g. 52.9 does not make any sense here - I don't think the "bias"
        #  correction of the timestamps (see the function 'load_data()' in 'data_processing.py'):
        #     time_range -= time_range[0]
        #  is corrected...
        #  EDIT: Maybe the elimination of zero-velocity data plays a role(?)

        observation_matrix = self._observation_matrix(self.robot_data_calibration,
                                                       start_index=self.robot_data_calibration.non_static_start_index,
                                                       end_index=self.robot_data_calibration.non_static_end_index)
        measurement_vector = self._measurement_vector(self.robot_data_calibration,
                                                       start_index=self.robot_data_calibration.non_static_start_index,
                                                       end_index=self.robot_data_calibration.non_static_end_index)

        # sklearn fit
        OLS = LinearRegression(fit_intercept=False)
        OLS.fit(observation_matrix, measurement_vector)
        measurement_vector_ols_estimation = OLS.predict(observation_matrix)
        n_samples_est = round(len(measurement_vector) / self.robot_dynamics.n_joints)

        # Reshape measurement vector from (n_joints*n_samples x 1) to (n_samples x n_joints)
        assert n_samples_est * self.robot_dynamics.n_joints == len(measurement_vector_ols_estimation) \
               == np.shape(observation_matrix)[0]
        measurement_vector_reshape = np.reshape(measurement_vector, (self.robot_dynamics.n_joints, n_samples_est))
        measurement_vector_ols_estimation_reshape = np.reshape(measurement_vector_ols_estimation,
                                                               (self.robot_dynamics.n_joints, n_samples_est))

        # Compute weights (the reciprocal of the estimated variance of the error)
        residuals = measurement_vector_reshape - measurement_vector_ols_estimation_reshape
        residual_sum_of_squares = np.sum(np.square(residuals), axis=1)
        variance_residual = residual_sum_of_squares / (
                    n_samples_est - np.sum(self.robot_dynamics.number_of_parameters()))

        standard_deviation = np.sqrt(variance_residual)
        wls_sample_weights = np.repeat(1 / standard_deviation, n_samples_est)

        # Weighted Least Squares solution
        wls_calibration = LinearRegression(fit_intercept=False)
        wls_calibration.fit(observation_matrix, measurement_vector, sample_weight=wls_sample_weights)
        y_pred = wls_calibration.predict(observation_matrix)

        y_pred_reshaped = np.reshape(y_pred, (self.robot_dynamics.n_joints, y_pred.shape[0] // self.robot_dynamics.n_joints))
        measurement_vector_reshaped = np.reshape(measurement_vector, (self.robot_dynamics.n_joints, y_pred.shape[0] // self.robot_dynamics.n_joints))
        assert y_pred_reshaped.shape == measurement_vector_reshaped.shape
        mse = RobotCalibration.get_mse(measurement_vector_reshaped, y_pred_reshaped)
        print(f"MSE: {mse}")
        weighted_observation_matrix = (wls_sample_weights*observation_matrix.T).T
        std_dev_parameter_estimate = np.sqrt(np.diagonal(np.linalg.inv(weighted_observation_matrix.T @ weighted_observation_matrix)))  # calculates the standard deviation of the parameter estimates from the diagonal elements (variance) of the covariance matrix
        cond = RobotCalibration._evaluate_dynamics_excitation_as_cost((wls_sample_weights*observation_matrix.T).T, metric="cond")
        print(f"Condition no. of weighted observation matrix: {cond}.")

        self.parameters = wls_calibration.coef_
        print(f"std_dev_parameter_estimate.shape: {std_dev_parameter_estimate.shape}")
        print(f"self.parameters.shape: {self.parameters.shape}")
        rel_std_dev_parameter_estimate = 100 * (std_dev_parameter_estimate / self.parameters)
        print(f"parameters: {self.robot_dynamics.parameters()}")
        print(f"rel_std_dev_parameter_estimate: {rel_std_dev_parameter_estimate}")

        # cache_numpy(from_cache(filename_parameters), lambda: parameters)
        cache_csv(from_cache(filename_parameters), lambda: self.parameters)
        return wls_calibration.coef_
        # return self.predict(self.robot_data_calibration, parameters, 'calibration_output')

    def predict(self, robot_data_predict, parameters, filename_predicted_output):

        def compute_prediction():
            observation_matrix = self._observation_matrix(robot_data_predict)
            estimated_output = observation_matrix @ parameters

            n_samples_ds = round(observation_matrix.shape[0] / self.robot_dynamics.n_joints)
            assert n_samples_ds * self.robot_dynamics.n_joints == observation_matrix.shape[0]
            output_predicted_reshaped = np.reshape(estimated_output, (self.robot_dynamics.n_joints, n_samples_ds))

            #return robot_data_predict.time[
            #       ::self.downsampling_factor], output_predicted_reshaped  # TODO: CORRECT ERROR; 'time' does not correspond in length to 'output_predicted_reshaped'
            return output_predicted_reshaped

        return cache_csv(from_cache(filename_predicted_output), compute_prediction)

    def _get_plot_values_for(self, data, parameters):
        observation_matrix = self._observation_matrix(data)
        n_samples = observation_matrix.shape[0] // self.robot_dynamics.n_joints
        estimated_data = np.reshape(observation_matrix @ parameters, (self.robot_dynamics.n_joints, n_samples))
        measured_data = np.reshape(self._measurement_vector(data),
                                              (self.robot_dynamics.n_joints, n_samples))
        #error = measured_data - estimated_data
        t_data = np.linspace(0, data.dt_nominal * self.downsampling_factor * n_samples, n_samples)
        return t_data, measured_data, estimated_data

    def plot_calibration(self, parameters):
        try:
            import matplotlib.colors
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings
            warnings.warn("The matplotlib package is not installed, please install it for plotting the calibration.")

        t, measured_output_reshaped, estimated_output_reshaped = self._get_plot_values_for(self.robot_data_calibration, parameters)
        error = measured_output_reshaped - estimated_output_reshaped

        fig = plt.figure()
        gs = fig.add_gridspec(2, 1, hspace=0.03)
        axs = gs.subplots(sharex='col', sharey='all')

        def darken_color(plot_color, darken_amount=0.35):
            """Computes rgb values to a darkened color"""
            line_color_rgb = matplotlib.colors.ColorConverter.to_rgb(plot_color)
            line_color_hsv = matplotlib.colors.rgb_to_hsv(line_color_rgb)
            darkened_line_color_hsv = line_color_hsv - np.array([0, 0, darken_amount])
            darkened_line_color_rgb = matplotlib.colors.hsv_to_rgb(darkened_line_color_hsv)
            return darkened_line_color_rgb

        # Current
        for j in range(self.robot_dynamics.n_joints):
            axs[0].plot(t, measured_output_reshaped[j, :], '-', color=plot_colors[j], linewidth=2.5,
                        label=f'joint {j + 1}, meas.')
            axs[0].plot(t, estimated_output_reshaped[j, :], color=darken_color(plot_colors[j]), linewidth=2.5,
                        label=f'joint {j + 1}, pred.')
            axs[0].legend(loc="best")
        axs[0].set_xlim([t[0], t[-1]])
        axs[0].set_title('Calibration')

        # Error
        for j in range(self.robot_dynamics.n_joints):
            axs[1].plot(t, error[j].T, '-', color=plot_colors[j], linewidth=1.3, label=f'joint {j + 1}')
            plt.legend(loc="upper left")
        axs[1].set_xlim([t[0], t[-1]])

        for ax in axs.flat:
            ax.label_outer()
        plt.setp(axs[0], ylabel='Current [A]')
        plt.setp(axs[1], ylabel='Error [A]')
        plt.setp(axs[1], xlabel='Time [s]')

        plt.show()

    def plot_prediction(self, filename_predict):

        t, y = load_object(from_cache(filename_predict + '.pickle'))
        # n_samples = self.robot_data_prediction.
        # t = np.linspace(0, self.robot_data_calibration.dt_nominal * n_samples, n_samples)

        observation_matrix = self._observation_matrix(self.robot_data_calibration)
        n_samples = observation_matrix.shape[0] // self.robot_dynamics.n_joints
        estimated_output_reshaped = np.reshape(observation_matrix @ parameters,
                                               (self.robot_dynamics.n_joints, n_samples))
        measured_output_reshaped = np.reshape(self._measurement_vector(self.robot_data_calibration),
                                              (self.robot_dynamics.n_joints, n_samples))
        error = measured_output_reshaped - estimated_output_reshaped
        t = np.linspace(0, self.robot_data_calibration.dt_nominal * self.downsampling_factor * n_samples, n_samples)

        import matplotlib.colors
        import matplotlib.pyplot as plt
        fig = plt.figure()
        gs = fig.add_gridspec(2, 1, hspace=0.03)
        axs = gs.subplots(sharex='col', sharey='all')

        def darken_color(plot_color, darken_amount=0.35):
            """Computes rgb values to a darkened color"""
            line_color_rgb = matplotlib.colors.ColorConverter.to_rgb(plot_color)
            line_color_hsv = matplotlib.colors.rgb_to_hsv(line_color_rgb)
            darkened_line_color_hsv = line_color_hsv - np.array([0, 0, darken_amount])
            darkened_line_color_rgb = matplotlib.colors.hsv_to_rgb(darkened_line_color_hsv)
            return darkened_line_color_rgb

        # Current
        for j in range(self.robot_dynamics.n_joints):
            axs[0].plot(t, measured_output_reshaped[j, :], '-', color=plot_colors[j], linewidth=1.5,
                        label=f'joint {j}, meas.')
            axs[0].plot(t, estimated_output_reshaped[j, :], color=darken_color(plot_colors[j]), linewidth=1,
                        label=f'joint {j}, pred.')
        axs[0].set_xlim([t[0], t[-1]])
        axs[0].set_title('Prediction')

        # Error
        for j in range(self.robot_dynamics.n_joints):
            axs[1].plot(t, error[j].T, '-', color=plot_colors[j], linewidth=1.3, label=f'joint {j + 1}')
        axs[1].set_xlim([t[0], t[-1]])

        for ax in axs.flat:
            ax.label_outer()
        plt.setp(axs[0], ylabel='Current [A]')
        plt.setp(axs[1], ylabel='Error [A]')
        plt.setp(axs[1], xlabel='Time [s]')

        plt.show()

    def plot_calibrate_and_validate(self, parameters):
        try:
            import matplotlib.pyplot as plt
            from matplotlib.ticker import MultipleLocator
            from matplotlib.colors import hsv_to_rgb, ColorConverter, rgb_to_hsv
        except ImportError:
            import warnings
            warnings.warn("The matplotlib package is not installed, please install it for plotting the calibration.")

        ## Calibration and Validation estimation and data
        t_calibration, measured_output_reshaped, estimated_output_reshaped = self._get_plot_values_for(self.robot_data_calibration, parameters)
        t_validation, measured_validation_output_reshaped, validation_output_reshaped = self._get_plot_values_for(self.robot_data_validation, parameters)
        t_validation += t_calibration[-1]

        def darken_color(plot_color, darken_amount=0.45):
            """Computes rgb values to a darkened color"""
            line_color_rgb = ColorConverter.to_rgb(plot_color)
            line_color_hsv = rgb_to_hsv(line_color_rgb)
            darkened_line_color_hsv = line_color_hsv - np.array([0, 0, darken_amount])
            darkened_line_color_rgb = hsv_to_rgb(darkened_line_color_hsv)
            return darkened_line_color_rgb


        fig = plt.figure()
        gs = fig.add_gridspec(2, 2, hspace=0.03, wspace=0,
                                width_ratios=[np.max(t_calibration) - np.min(t_calibration),
                                            np.max(t_validation) - np.min(t_validation)])
        axs = gs.subplots(sharex='col', sharey='all')


        linewidth_meas = 2.5
        linewidth_est = 2.0
        linetype_meas = '-'
        linetype_est = '--'

        # Estimation data - current
        for j in range(self.robot_dynamics.n_joints):
            axs[0, 0].plot(t_calibration, measured_output_reshaped[j, :].T, linetype_meas, color=plot_colors[j], linewidth=linewidth_meas,
                            label=f'joint {j+1}, meas.')
            axs[0, 0].plot(t_calibration, estimated_output_reshaped[j, :].T, linetype_est, color=darken_color(plot_colors[j]), linewidth=linewidth_est, label=f'joint {j+1}, est.')
        axs[0, 0].set_xlim([t_calibration[0], t_calibration[-1]])
        axs[0, 0].set_title('Calibration')

        # Validation data - current
        mse = self.get_mse(measured_validation_output_reshaped, validation_output_reshaped)
        for j in range(self.robot_dynamics.n_joints):
            axs[0, 1].plot(t_validation, measured_validation_output_reshaped[j, :].T, linetype_meas, color=plot_colors[j], linewidth=linewidth_meas)
            axs[0, 1].plot(t_validation, validation_output_reshaped[j, :].T, linetype_est, color=darken_color(plot_colors[j]), linewidth=linewidth_est)
        axs[0, 1].set_xlim([t_validation[0], t_validation[-1]])
        axs[0, 1].set_title('Validation')

        # Estimation data - error
        error_est = (measured_output_reshaped - estimated_output_reshaped)
        for j in range(self.robot_dynamics.n_joints):
            axs[1, 0].plot(t_calibration, error_est[j].T, linetype_meas, color=plot_colors[j], linewidth=linewidth_meas, label=f'joint {j + 1}')
        axs[1, 0].set_xlim([t_calibration[0], t_calibration[-1]])

        # Validation data - error
        error_val = (measured_validation_output_reshaped - validation_output_reshaped)
        for j in range(self.robot_dynamics.n_joints):
            axs[1, 1].plot(t_validation, error_val[j].T, linetype_meas, color=plot_colors[j], linewidth=linewidth_meas, label=f'joint {j + 1}')
        axs[1, 1].set_xlim([t_validation[0], t_validation[-1]])

        # equate xtick spacing of right plots to those of left plots
        xticks_diff = axs[1, 0].get_xticks()[1] - axs[1, 0].get_xticks()[0]
        axs[1, 0].xaxis.set_major_locator(MultipleLocator(xticks_diff))
        axs[1, 1].xaxis.set_major_locator(MultipleLocator(xticks_diff))

        for ax in axs.flat:
            ax.label_outer()
        plt.setp(axs[0, 0], ylabel='Current [A]')
        plt.setp(axs[1, 0], ylabel='Error [A]')
        
        pos_label = ((len(t_validation) / len(t_calibration)) + 1) / 2
        axs[1,0].set_xlabel("Time [s]", fontsize='large', ha="center", position=(pos_label,pos_label))

        # Legend position
        axs[0, 0].legend(loc='lower left', ncol=self.robot_dynamics.n_joints)
        axs[1, 0].legend(loc="upper left", ncol=self.robot_dynamics.n_joints)
        plt.show()
    
    @staticmethod
    def _evaluate_dynamics_excitation_as_cost(observation_matrix, metric="cond"):
        assert metric in {"cond", "determinant", "log_determinant", "minimum_singular_value"}

        if metric == "cond":
            cost = np.linalg.cond(observation_matrix)
        elif metric == "determinant":
            cost = np.linalg.det(observation_matrix.T @ observation_matrix)
        elif metric == "log_determinant":
            cost = np.log(np.linalg.det(observation_matrix.T @ observation_matrix))
        elif metric == "minimum_singular_value":
            cost = np.linalg.svd(observation_matrix.T @ observation_matrix)  # validate this
        else:
            cost = 0
        return cost

    @staticmethod
    def _downsample(y, downsampling_factor):
        """The decimate procedure down-samples the signal such that the matrix system (that is later to be inverted) is not
        larger than strictly required. The signal.decimate() function can also low-pass filter the signal before
        down-sampling, but for IIR filters unfortunately only the Chebyshev filter is available which has (unwanted) ripple
        in the passband unlike the Butterworth filter that we use. The approach for downsampling is simply picking every
        downsampling_factor'th sample of the data."""

        y_ds = y[::downsampling_factor, :]
        return y_ds

    @staticmethod
    def _parallel_filter(y, dt, f_dyn):
        """Applies a 4th order Butterworth (IIR) filter for each row in y having a cutoff frequency of 2*f_dyn."""
        # Link to cut-off freq. eq.: https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=6151858
        parallel_filter_order = 4
        cutoff_freq_parallel = 2 * f_dyn
        parallel_filter = signal.butter(parallel_filter_order, cutoff_freq_parallel, btype='low', output='sos',
                                        fs=1 / dt)

        y_pf = signal.sosfiltfilt(parallel_filter, y, axis=0)
        return y_pf

    @staticmethod
    def _trajectory_filtering_and_central_difference(q_m, dt, f_dyn, idx_start=None, idx_end=None):
        assert idx_end < q_m.shape[1], "'idx_end' must not be greater than the dataset size."

        trajectory_filter_order = 4
        cutoff_freq_trajectory = 5 * f_dyn  # Cut-off frequency should be around 5*f_dyn = 50 Hz(?)
        trajectory_filter = signal.butter(trajectory_filter_order, cutoff_freq_trajectory, btype='low', output='sos',
                                          fs=1 / dt)
        q_tf = signal.sosfiltfilt(trajectory_filter, q_m, axis=1)

        # Obtain first and second order time-derivatives of measured and filtered trajectory
        q_tf, qd_tf, qdd_tf = central_finite_difference(q_tf, dt, order=2)

        # Truncate/crop data
        # assert q_tf.shape[1] > idx_end + 1, f"'idx_end' must be smaller than or equal to the length of the signal along axis 1"
        q_tf = q_tf[:, idx_start:idx_end]
        qd_tf = qd_tf[:, idx_start:idx_end]
        qdd_tf = qdd_tf[:, idx_start:idx_end]

        assert q_tf.shape == qd_tf.shape == qdd_tf.shape, f"q_tf.shape == {q_tf.shape}, qd_tf.shape == {qd_tf.shape}, qdd_tf.shape == {qdd_tf.shape}"

        return q_tf, qd_tf, qdd_tf

    @staticmethod
    def get_mse(y_meas, y_pred):
        """
        This function calculates the mean squared error of each channel in y using sklearn.metrics.mean_squared_error.
        """
        assert y_meas.shape == y_pred.shape

        n_channels = y_meas.shape[0]  # no. of channels
        mse = np.zeros(n_channels)
        for i in range(n_channels):
            mse[i] = mean_squared_error(y_meas[i, :], y_pred[i, :])
        return mse


import os
import sys
import pickle
import numpy as np
from pathlib import Path
import pandas as pd


def project_root() -> Path:
    return Path(__file__).parent.parent


def from_project_root(filepath):
    return project_root().joinpath(filepath)


def from_cache(filepath):
    new_dir = project_root().joinpath('cache')
    if not Path(new_dir).is_dir():
        os.mkdir(new_dir)
    return str(project_root().joinpath('cache', filepath))


def safe_open(filepath, mode='r'):
    """
    Checks if file exists and gives a friendly error in case it doesn't.
    """
    if not os.path.isfile(filepath):
        print(f"File {filepath} not found. Current dir is {os.getcwd()}")
        sys.exit(1)
    return open(filepath, mode)


def store_object(file, obj):
    with open(file, "wb") as f:
        pickle.dump(obj, f)


def load_object(file):
    with safe_open(file, "rb") as f:
        return pickle.load(f)


def cache_object(filename, callable):
    """
    Caches the return value of the callable using pickle.
    """
    file = filename + '.pickle'
    if os.path.exists(file):
        return load_object(file)
    else:
        obj = callable()
        store_object(file, obj)
        return obj


def store_numpy(file, nparr):
    assert file[-4:] == '.npy'
    # assert isinstance(nparr, (np.ndarray, np.generic))
    print(f"file: {file}")
    print(f"nparr: {nparr}")
    with open(file, "wb") as f:
        np.save(f, nparr)

def store_csv(file, data):
    assert file[-4:] == '.csv'
    df = pd.DataFrame(data)
    df.to_csv(file, index=False, header=False)

def load_numpy(file):
    with safe_open(file, "rb") as f:
        return np.load(f)

def load_csv(file):
    return pd.read_csv(file, header=None)

def cache_numpy(file, callable):
    """
        Caches the return value of the callable using numpy.
    """
    if not file[-4:] == '.npy':
        file = file + '.npy'

    if os.path.exists(file):
        return load_numpy(file)
    else:
        expr = callable()
        store_numpy(file, expr)
        return expr



def cache_csv(file, callable):
    if not file[-4:] == '.csv':
        file = file + '.csv'

    if os.path.exists(file):
        return load_csv(file)
    else:
        expr = callable()
        store_csv(file, expr)
        return expr


import numpy as np


def central_finite_difference(x, dt, order):
    """Computes the first and (optionally) second-order time-derivatives alon axis 1. The dimension of the output is
    equal to the dimension of the input minus 2."""
    assert order in {1, 2}

    xd = np.gradient(x, dt, axis=1)

    if order == 2:
        xdd = (x[:, 2:] - 2 * x[:, 1:-1] + x[:, :-2]) / (dt**2)
        assert x[:, 1:-1].shape == xd[:, 1:-1].shape == xdd.shape
        return x[:, 1:-1], xd[:, 1:-1], xdd

    assert x[:, 1:-1].shape == xd[:, 1:-1].shape

    return x[:, 1:-1], xd[:, 1:-1]


def crop(x, idx_start=None, idx_end=None):
    assert x.shape[1] > idx_end, f"'idx_end' must be smaller than the length of the signal along axis 1"

    if idx_end is not None:
        x = x[:, :idx_end]
    if idx_start is not None:
        x = x[:, idx_start:]

    return x



import math
import numpy as np
import sympy as sp


# Numerical computation layer

def npcross(a, b):
    assert a.shape == (3, 1), a
    assert b.shape == (3, 1), b
    res = np.cross(a, b, axisa=0, axisb=0).transpose()
    assert res.shape == (3, 1)
    return res


def npdot(a, b):
    (n1, m1) = a.shape
    (n2, m2) = b.shape
    res = np.dot(a, b)
    assert res.shape == (n1, m2), a.shape
    return res


def npvector(l): return np.array(l).transpose().reshape((len(l), 1))


def npmatrix(l): return np.array(l)


def npzeros_array(n1): return np.zeros(n1)


def npzeros_matrix(n1, n2): return np.zeros((n1, n2))


def npcos(n): return math.cos(n)


def npsin(n): return math.sin(n)


def npeye(n): return np.eye(n)

# Symbolic computation layer

def spcross(v1, v2):
    assert v1.shape == (3, 1), v1
    assert v2.shape == (3, 1), v2
    res = v1.transpose().cross(v2.transpose()).transpose()
    assert res.shape == (3, 1)
    return res


def spdot(a, b):
    (n1, m1) = a.shape
    (n2, m2) = b.shape
    res = a * b
    assert res.shape == (n1, m2)
    return res


def spvector(l): return sp.Matrix(l)


def spmatrix(l): return sp.Matrix(l)


def spzeros_array(n1): return sp.zeros(rows=n1, cols=1)


def spzeros_matrix(n1, n2): return sp.zeros(n1, n2)


def spcos(n): return sp.cos(n)


def spsin(n): return sp.sin(n)


def speye(n): return sp.eye(n)


from math import pi
from dataclasses import dataclass
import pandas as pd
import sympy as sp


plot_colors = ['red', 'green', 'blue', 'chocolate', 'crimson', 'fuchsia', 'indigo', 'orange']

JOINT_N = "#"


@dataclass
class ModifiedDH:
    d: list
    a: list
    alpha: list
    q: list
    n_joints: int

    def __init__(self, d, a, alpha, q) -> None:
        self.d = d
        self.a = a
        self.alpha = alpha
        self.q = q
        assert len(self.d) == len(self.a) == len(self.alpha)
        self.set_n_joints()
    
    def set_n_joints(self) -> int:
        self.n_joints = len(self.d)-1


def convert_file_to_mdh(filename):
    if filename[-3:] == "csv":
        df = pd.read_csv(filename)
    df = df.fillna(value='None')
    d = [float(d) if d != 'None' else None for d in df.d]
    a = [float(a) if a != 'None' else None for a in df.a]
    alpha = []
    for alpha_i in df.alpha:
        alpha.append(input_with_pi_to_float(alpha_i))
    # insert the extra joint at index 0
    d.insert(0,float(0))
    a.insert(0,float(0))
    alpha.insert(0,float(0))
    mdh = ModifiedDH(d,a,alpha,None) # TODO: add q to the listing, and remove njoints, remove sp.integer(0)
    q = [sp.Integer(0)] + [sp.symbols(f"q{j}") for j in range(1, mdh.n_joints + 1)]
    mdh.q = q
    return mdh

def input_with_pi_to_float(input):
    if input == 'None':
        return None
    elif isinstance(input, str) and "pi" in input:
        input = input.split("/")
        # Case: no / in input, i.e. either pi or a number, or -pi
        if len(input) == 1: # this means no / is in input
            if "pi" == input[0]:
                return pi
            elif "-pi" == input[0]:
                return -pi
            else:
                return float(input)
        # Case: / in input, either pi/2, pi/4, -pi/2, -pi/8, or number/number
        elif len(input) == 2:
            if "-pi" == input[0]:
                return -pi/float(input[1])
            elif "pi" == input[0]:
                return pi/float(input[1])
        else:
            print(f"Whoops, len of input is greater than 2: {len(input)}")
    else:
        return float(input)




import sympy as sp


def sym_mat_to_subs(sym_mats, num_mats):
    subs = {}

    for s_mat, n_mat in zip(sym_mats, num_mats):
        subs = {**subs, **{s: v for s, v in zip(s_mat, n_mat) if s != 0}}

    return subs


def replace_first_moments(args):

    tau_sym_j = args[0][0]  # tau_sym[j]
    j = args[0][1]
    m = args[1]  # [m1, ..., mN]
    pc = args[2]  # PC = [PC[1], ..., PC[N]], PC[j] = [PCxj, PCyj, PCzj]
    m_pc = args[3]  # mPC = [mPC[1], ..., mPC[N]], mPC[j] = [mXj, mYj, mZj]
    n_joints = args[4]

    # TODO: REMOVE THE 'IGNORE J=0' STUFF
    if j == 0:
        print(f"Ignore joint {j}...")
        return tau_sym_j

    n_cartesian = len(m_pc[0])
    for jj in range(j, n_joints + 1):  # The joint j torque equation is affected by dynamic coefficients only for links jj >= j
        for i in range(n_cartesian):  # Cartesian coordinates loop
            # Print progression counter
            total = (n_joints + 1 - j) * n_cartesian
            progress = (jj - j) * n_cartesian + i + 1
            print(f"Task {j}: {progress}/{total} (tau[{j}]: {m[jj]}*{pc[jj][i]} -> {m_pc[jj][i]})")
            tau_sym_j = tau_sym_j.expand().subs(m[jj] * pc[jj][i], m_pc[jj][i])  # expand() is - for unknown reasons - needed for subs() to consistently detect the "m*PC" products

    return tau_sym_j


def compute_regressor_row(args): 
    """
    We compute the regressor via symbolic differentiation. Each torque equation must be linearizable with respect to
    the dynamic coefficients.

    Example:
        Given the equation 'tau = sin(q)*a', tau is linearizable with respect to 'a', and the regressor 'sin(q)' can be
        obtained by partial differentiation of 'tau' with respect to 'a'.
    """
    tau_sym_linearizable_j = args[0][0]  # tau_sym_linearizable[j]
    j = args[0][1]
    n_joints = args[1]
    p_linear = args[2]

    n_par = [len(p_linear[j]) for j in range(len(p_linear))]
    reg_row_j = sp.zeros(1, sum(n_par))  # Initialization

    if j == 0:
        print(f"Ignore joint {j}...")
        return reg_row_j

    # For joint j, we loop through the parameters belonging to joints/links >= j. This is because it is physically
    # impossible for torque eq. j to include a parameter related to proximal links (< j). We divide the parameter
    # loop (for joints >= j) in two variables 'jj' and 'i':
    #   'jj' describes the joint, which the parameter belongs to
    #   'i'  describes the parameter's index/number for joint jj.
    for jj in range(j, n_joints + 1):  # Joint loop including this and distal (succeeding) joints (jj >= j)
        for i in range(n_par[jj]):  # joint jj parameter loop
            column_idx = sum(n_par[:jj]) + i
            print(
                f"Computing regressor(row={j}/{n_joints + 1}, column={column_idx + 1}/{sum(n_par)}) by analyzing dependency of tau[{j}] on joint {jj}'s parameter {i}: {p_linear[jj][i]}")
            reg_row_j[0, column_idx] = sp.diff(tau_sym_linearizable_j, p_linear[jj][i])

    return reg_row_j



import numpy as np
import sympy as sp
from multiprocessing import Pool
from itertools import product, chain
import sys

from aurt.file_system import cache_object, from_cache
from aurt.dynamics_aux import sym_mat_to_subs, replace_first_moments, compute_regressor_row
from aurt.num_sym_layers import spvector, spcross, spdot
from aurt.data_processing import ModifiedDH


class RigidBodyDynamics:
    __qr_numerical_threshold = 1e-12  # Numerical threshold used in identifying the base inertial parameters
    __max_number_of_rank_evaluations = 100  # Max. no. of rank evaluations in computing the base inertial parameters
    __n_regressor_evals_per_rank_calculation = 2  # No. of regressor evaluations per rank calculation
    __min_rank_evals = 4

    def __init__(self, modified_dh: ModifiedDH, gravity=None, tcp_force_torque=None):
        self.mdh = modified_dh
        self.n_joints = modified_dh.n_joints

        self.q = self.mdh.q
        self.qd = [sp.Integer(0)] + [sp.symbols(f"qd{j}") for j in range(1, self.n_joints + 1)]
        self.qdd = [sp.Integer(0)] + [sp.symbols(f"qdd{j}") for j in range(1, self.n_joints + 1)]

        self.__m = sp.symbols([f"m{j}" for j in range(self.n_joints + 1)])
        self.__mX = sp.symbols([f"mX{j}" for j in range(self.n_joints + 1)])
        self.__mY = sp.symbols([f"mY{j}" for j in range(self.n_joints + 1)])
        self.__mZ = sp.symbols([f"mZ{j}" for j in range(self.n_joints + 1)])

        # TODO: clean code
        PC = [spvector([0.0, 0.0, 0.0])] * (self.n_joints + 1)
        for j in range(1, self.n_joints + 1):
            PC[j] = spvector([sp.symbols(f"PCx{j}"), sp.symbols(f"PCy{j}"), sp.symbols(f"PCz{j}")])
        self.__pc = PC

        self.__m_pc = [[self.__mX[j], self.__mY[j], self.__mZ[j]] for j in range(self.n_joints + 1)]

        self.__XX = sp.symbols([f"XX{j}" for j in range(self.n_joints + 1)])
        self.__XY = sp.symbols([f"XY{j}" for j in range(self.n_joints + 1)])
        self.__XZ = sp.symbols([f"XZ{j}" for j in range(self.n_joints + 1)])
        self.__YY = sp.symbols([f"YY{j}" for j in range(self.n_joints + 1)])
        self.__YZ = sp.symbols([f"YZ{j}" for j in range(self.n_joints + 1)])
        self.__ZZ = sp.symbols([f"ZZ{j}" for j in range(self.n_joints + 1)])
        self.__i_cor = [sp.zeros(3, 3) for _ in range(self.n_joints + 1)]
        for j in range(self.n_joints + 1):
            self.__i_cor[j] = sp.Matrix([
                [self.__XX[j], self.__XY[j], self.__XZ[j]],
                [self.__XY[j], self.__YY[j], self.__YZ[j]],
                [self.__XZ[j], self.__YZ[j], self.__ZZ[j]]
            ])

        if gravity is None:
            print(f"No gravity direction was specified. Assuming the first joint axis of rotation to be parallel to gravity...")
            gravity = [0, 0, -9.81]
        self.gravity = gravity
        gx, gy, gz = sp.symbols(f"gx gy gz")
        self.__g = sp.Matrix([gx, gy, gz])

        fx, fy, fz, nx, ny, nz = sp.symbols(f"fx fy fz nx ny nz")
        self.__f_tcp = sp.Matrix([fx, fy, fz])  # Force at the TCP
        self.__n_tcp = sp.Matrix([nx, ny, nz])  # Moment at the TCP

        if tcp_force_torque is None:
            f_tcp_num = np.array([0, 0, 0])
            n_tcp_num = np.array([0, 0, 0])
            tcp_force_torque = [f_tcp_num, n_tcp_num]
        self.__f_tcp_num = tcp_force_torque[0]
        self.__n_tcp_num = tcp_force_torque[1]

        # Filepaths
        #self.filepath_dynamics = from_cache('rigid_body_dynamics')
        #self.__filepath_regressor_joint = from_cache('rigid_body_dynamics_regressor_joint_')

        self.n_params = None
        self.params = None
        self.regressor_linear = None
        self.idx_base_exist = None

    def filepath_regressor_joint(self, j):
        return from_cache(f'rigid_body_dynamics_regressor_joint_{j}')

    @property
    def n_params(self):
        if self._n_params is None:
            self.number_of_parameters()
        return self._n_params

    @property
    def params(self):
        if self._params is None:
            self.parameters()
        return self._params

    @n_params.setter
    def n_params(self, val):
        self._n_params = val

    @params.setter
    def params(self, val):
        self._params = val

    def parameters(self):
        """
        Returns a list of 'n_joints + 1' elements with each element comprising a list of all rigid body parameters
        related to that corresponding link.
        """
        base_params_information = self.__base_parameters_information()
        self.params = base_params_information[2][1:]
        self.n_params = base_params_information[1][1:]
        return self.params

    def number_of_parameters(self):
        self.n_params = self.__base_parameters_information()[1][1:]
        return self.n_params

    def __parameters_linear(self):
        """
        Returns a list of 'n_joints + 1' elements with each element comprising a list of all parameters related to that
        corresponding rigid body.
        """
        return [[self.__XX[j], self.__XY[j], self.__XZ[j], self.__YY[j], self.__YZ[j], self.__ZZ[j],
                 self.__mX[j], self.__mY[j], self.__mZ[j], self.__m[j]] for j in range(self.n_joints + 1)]

    def __regressor_linear(self):
        if self.regressor_linear is not None:
            return self.regressor_linear

        dynamics_linearizable = self.dynamics()

        js = list(range(self.n_joints + 1))
        tau_per_task = [dynamics_linearizable[j] for j in js]  # Allows one to control how many tasks by controlling how many js
        data_per_task = list(product(zip(tau_per_task, js), [self.n_joints], [self.__parameters_linear()]))

        with Pool() as p:
            reg = p.map(compute_regressor_row, data_per_task)

        res = sp.Matrix(reg)
        self.regressor_linear = res
        return res

    def __regressor_linear_exist(self):
        # In the eq. for joint j, the dynamic parameters of proximal links (joints < j, i.e. closer to the base) will never
        # exist, i.e. the dynamic parameter of joint j will not be part of the equations for joints > j.
        regressor = self.__regressor_linear()
        idx_linear_exist = self.__parameters_linear_exist(regressor)[0]

        # Removes zero columns of regressor corresponding to parameters with no influence on dynamics
        idx_linear_exist_global = np.where(list(chain.from_iterable(idx_linear_exist)))[0].tolist()

        return regressor[:, idx_linear_exist_global]

    def __parameters_linear_exist(self, regressor):
        # In the regressor, identify the zero columns corresponding to parameters which does not matter to the system.
        number_of_parameters_linear_model = [len(self.__parameters_linear()[j]) for j in range(len(self.__parameters_linear()))]

        # Initialization
        idx_linear_exist = [[True for _ in range(number_of_parameters_linear_model[j])] for j in range(len(self.__parameters_linear()))]
        n_par_linear_exist = number_of_parameters_linear_model.copy()
        p_linear_exist = self.__parameters_linear().copy()

        # Computing parameter existence
        for j in range(self.n_joints + 1):
            for i in reversed(range(number_of_parameters_linear_model[j])):  # parameter loop
                idx_column = sum(number_of_parameters_linear_model[:j]) + i
                if not any(regressor[:, idx_column]):
                    idx_linear_exist[j][i] = False
                    n_par_linear_exist[j] -= 1
                    del p_linear_exist[j][i]

        return idx_linear_exist, n_par_linear_exist, p_linear_exist

    def __base_parameters_information(self):
        """Identifies the base parameters of the robot dynamics."""
        def compute_base_parameters_information():
            args_sym = self.q[1:] + self.qd[1:] + self.qdd[1:]
            sys.setrecursionlimit(int(1e6))  # Prevents errors in sympy lambdify
            regressor_with_instantiated_parameters_func = sp.lambdify(args_sym, cache_object(
                from_cache('rigid_body_dynamics_regressor_with_instantiated_parameters'),
                lambda: self.__regressor_linear_with_instantiated_parameters()))
            idx_base_global = self.__indices_base_exist(regressor_with_instantiated_parameters_func)

            idx_linear_exist, n_par_linear_exist, p_linear_exist = self.__parameters_linear_exist(self.__regressor_linear())
            p_linear_exist_vector = sp.Matrix(list(chain.from_iterable(p_linear_exist)))  # flatten 2D list to 1D list and convert 1D list to sympy.Matrix object

            p_base_vector = cache_object(from_cache('rigid_body_dynamics_base_parameters_vector'), lambda: p_linear_exist_vector[idx_base_global, :])

            # Initialization
            n_par_base = n_par_linear_exist.copy()
            p_base = p_linear_exist.copy()
            idx_is_base = [[True for _ in range(len(p_base[j]))] for j in range(len(p_base))]

            for j in range(self.n_joints + 1):
                for i in reversed(range(n_par_linear_exist[j])):
                    if not p_base[j][i] in p_base_vector.free_symbols:
                        idx_is_base[j][i] = False
                        n_par_base[j] -= 1
                        del p_base[j][i]

            assert np.count_nonzero(list(chain.from_iterable(idx_is_base))) == sum(n_par_base) == len(
                list(chain.from_iterable(p_base)))

            return idx_is_base, n_par_base, p_base

        idx_is_base, n_par_base, p_base = cache_object(from_cache('rigid_body_dynamics_parameter_information'),
                                                       compute_base_parameters_information)

        return idx_is_base, n_par_base, p_base

    def __numerical_alpha_to_symbolical_pi(self):
        alpha_sym = []
        for i in range(len(self.mdh.alpha)):
            pi_factor = self.mdh.alpha[i] / (np.pi / 2)
            if abs(round(pi_factor) - pi_factor) < 1e-2:
                alpha_sym.append(round(pi_factor)*sp.pi/2)
            else:
                alpha_sym.append(sp.symbols(f"alpha{i}"))
        return alpha_sym

    def __mdh_num_to_sym(self):
        d = [sp.symbols(f"d{i}") if d != 0 else sp.Integer(0) for i, d in enumerate(self.mdh.d)]
        a = [sp.symbols(f"a{i}") if a != 0 else sp.Integer(0) for i, a in enumerate(self.mdh.a)]
        alpha = self.__numerical_alpha_to_symbolical_pi()
        m = [0] * (self.n_joints + 1)
        for j in range(1, self.n_joints + 1):
            m[j] = sp.symbols(f"m{j}")
        return m, d, a, alpha

    def __regressor_linear_with_instantiated_parameters(self):
        _, d, a, _ = self.__mdh_num_to_sym()

        def load_regressor_and_subs():
            regressor_reduced = self.__regressor_linear_exist()
            return regressor_reduced.subs(
                sym_mat_to_subs([a, d, self.__g, self.__f_tcp, self.__n_tcp],
                                [self.mdh.a, self.mdh.d, self.gravity, self.__f_tcp_num, self.__n_tcp_num]))

        return load_regressor_and_subs()

    def __indices_base_exist(self, regressor_with_instantiated_parameters):
        """This function computes the indices for the base parameters. The number of base parameters is obtained as the
        maximum obtainable rank of the observation matrix using a set of randomly generated dummy observations for the robot
        trajectory. The specific indices for the base parameters are obtained by conducting a QR decomposition of the
        observation matrix and analyzing the diagonal elements of the upper triangular (R) matrix."""
        # TODO: Change code such that 'dummy_pos, 'dummy_vel', and 'dummy_acc' can be called with a 2D list of
        #  indices to produce a 2D list of 'dummy_pos', 'dummy_vel', and 'dummy_acc'. This way, the regressor can be
        #  called with a 2D np.array(), which will drastically speed up the computation of the dummy observation matrix.

        if self.idx_base_exist is not None:
            return self.idx_base_exist

        # The set of dummy observations
        dummy_pos = np.array([-np.pi, -0.5 * np.pi, -0.25 * np.pi, 0.0, 0.25 * np.pi, 0.5 * np.pi, np.pi])
        dummy_vel = np.array([-1.0, -0.5, -0.25, 0.0, 0.25, 0.5, 1.0])
        dummy_acc = np.array([-1.0, -0.5, -0.25, 0.0, 0.25, 0.5, 1.0])
        assert len(dummy_pos) == len(dummy_vel) == len(dummy_acc)

        n_rank_evals = 0  # Loop counter for observation matrix concatenations
        rank_W = []

        random_idx_init = [[np.random.randint(self.n_joints + 1) for _ in range(self.n_joints)] for _ in range(3)]
        dummy_args_init = np.concatenate((dummy_pos[random_idx_init[0][:]], dummy_vel[random_idx_init[1][:]],
                                          dummy_acc[random_idx_init[2][:]]))

        W_N = regressor_with_instantiated_parameters(*dummy_args_init)
        W = regressor_with_instantiated_parameters(*dummy_args_init)

        # Continue the computations while the rank of the observation matrix 'W' keeps improving
        while n_rank_evals < RigidBodyDynamics.__min_rank_evals or rank_W[-1] > rank_W[-3]:  # While the rank of the observation matrix keeps increasing
            # Generate random indices for the dummy observations
            random_idx = [[[np.random.randint(self.n_joints + 1) for _ in range(self.n_joints)]
                           for _ in range(RigidBodyDynamics.__n_regressor_evals_per_rank_calculation)] for _ in range(3)]

            # Evaluate the regressor in a number of dummy observations and vertically stack the regressor matrices
            for i in range(RigidBodyDynamics.__n_regressor_evals_per_rank_calculation):
                dummy_args = np.concatenate(
                    (dummy_pos[random_idx[0][i][:]], dummy_vel[random_idx[1][i][:]], dummy_acc[random_idx[2][i][:]]))

                reg_i = regressor_with_instantiated_parameters(*dummy_args)  # Each index contains a (n_joints x n_par) regressor matrix
                W_N = np.append(W_N, reg_i, axis=0)
            W = np.append(W, W_N, axis=0)

            # Evaluate rank of observation matrix
            rank_W.append(np.linalg.matrix_rank(W))
            print(f"Rank of observation matrix: {rank_W[-1]}")

            n_rank_evals += 1
            if n_rank_evals > RigidBodyDynamics.__max_number_of_rank_evaluations:
                raise Exception(f"Numerical estimation of the number of base inertial parameters did not converge within {n_rank_evals * RigidBodyDynamics.__n_regressor_evals_per_rank_calculation} regressor evaluations.")

        r = np.linalg.qr(W, mode='r')
        idx_is_base = abs(np.diag(r)) > RigidBodyDynamics.__qr_numerical_threshold
        idx_base = np.where(idx_is_base)[0].tolist()
        assert len(idx_base) == rank_W[-1]

        self.idx_base_exist = idx_base
        return idx_base

    # def observation_matrix_joint_parameters_for_joint(self, j, j_par, q_num, qd_num, qdd_num):
    #     assert q_num.shape == qd_num.shape == qdd_num.shape
    #     assert 0 <= j < self.n_joints
    #     assert 0 <= j_par < self.n_joints

    #     regressor_j_jpar = self.regressor_joint_parameters_for_joint(j, j_par)
    #     args_sym = self.q[1:] + self.qd[1:] + self.qdd[1:]
    #     nonzeros = [not elem.is_zero for elem in regressor_j_jpar]
    #     sys.setrecursionlimit(int(1e6))
    #     regressor_j_jpar_nonzeros_fcn = sp.lambdify(args_sym, regressor_j_jpar[:, nonzeros], 'numpy')

    #     n_samples = q_num.shape[1]
    #     observation_matrix_j = np.zeros((n_samples, regressor_j_jpar.shape[1]))
    #     args_num = np.concatenate((q_num, qd_num, qdd_num))
    #     observation_matrix_j[:, nonzeros] = regressor_j_jpar_nonzeros_fcn(*args_num).transpose().squeeze(axis=2)
    #     return observation_matrix_j

    def observation_matrix_joint(self, j, q_num, qd_num, qdd_num):
        assert q_num.shape == qd_num.shape == qdd_num.shape
        assert 0 <= j < self.n_joints

        n_samples = q_num.shape[1]
        regressor_j = self.regressor_joint(j)
        sys.setrecursionlimit(int(1e6))
        nonzeros = [not elem.is_zero for elem in regressor_j]
        args_sym = self.q[1:] + self.qd[1:] + self.qdd[1:]
        observation_matrix_j = np.zeros((n_samples, regressor_j.shape[1]))
        regressor_j_nonzeros_fcn = sp.lambdify(args_sym, regressor_j[:, nonzeros], 'numpy')
        args_num = np.concatenate((q_num, qd_num, qdd_num))
        observation_matrix_j[:, nonzeros] = regressor_j_nonzeros_fcn(*args_num).squeeze().transpose()
        return observation_matrix_j

    # def observation_matrix(self, q_num, qd_num, qdd_num):
    #     assert q_num.shape == qd_num.shape == qdd_num.shape

    #     return np.vstack([self.observation_matrix_joint(j, q_num, qd_num, qdd_num) for j in range(self.n_joints)])

    def regressor_joint_parameters_for_joint(self, j, par_j):
        column_idx_start = sum(self.number_of_parameters()[:par_j])
        column_idx_end = column_idx_start + self.number_of_parameters()[par_j]
        res = self.regressor_joint(j)[:, column_idx_start:column_idx_end]
        return res

    def regressor_joint(self, j):
        return cache_object(self.filepath_regressor_joint(j+1), lambda: self.regressor()[j, :])

    def regressor(self, output_filename="rigid_body_dynamics_regressor"):
        filepath_regressor = from_cache(output_filename)

        def compute_regressor():
            regressor_linear_exist = self.__regressor_linear_with_instantiated_parameters()

            args_sym = self.q[1:] + self.qd[1:] + self.qdd[1:]  # list concatenation
            sys.setrecursionlimit(int(1e6))  # Prevents errors in sympy lambdify
            regressor_linear_exist_func = sp.lambdify(args_sym, regressor_linear_exist, 'numpy')

            parameter_indices_base = self.__indices_base_exist(regressor_linear_exist_func)

            for j in range(self.n_joints):
                cache_object(self.filepath_regressor_joint(j+1), lambda: regressor_linear_exist[j+1, parameter_indices_base])
            
            return regressor_linear_exist[1:, parameter_indices_base]

        return cache_object(filepath_regressor, compute_regressor)

    def dynamics(self):
        def compute_dynamics_and_replace_first_moments():
            rbd = self.__rigid_body_dynamics()
            js = list(range(self.n_joints + 1))
            dynamics_per_task = [rbd[j] for j in js]  # Allows one to control how many tasks by controlling how many js.
            data_per_task = list(product(zip(dynamics_per_task, js), [self.__m], [self.__pc], [self.__m_pc], [self.n_joints]))

            with Pool() as p:
                rbd_lin = p.map(replace_first_moments, data_per_task)
            return rbd_lin

        return cache_object(from_cache('rigid_body_dynamics'), compute_dynamics_and_replace_first_moments)

    def __get_P(self, a, d, alpha):
        """
        P[i] means the position of frame i wrt to i-1
        P[2] = [d, -0.2]^T
        P[5] = [a[5-1], 0, d[5]] means the position of frame 5 wrt to 4
        """
        P = [spvector([0, 0, 0]) for i in range(0, self.n_joints + 2)]

        for i in range(1, self.n_joints + 1):
            P[i] = spvector([a[i - 1], -sp.sin(alpha[i - 1]) * d[i], sp.cos(alpha[i - 1]) * d[i]])

        return P

    def __get_forward_kinematics(self, alpha):
        c = lambda i: sp.cos(self.q[i])
        s = lambda i: sp.sin(self.q[i])

        R_i_im1 = [sp.zeros(3, 3) for i in range(self.n_joints + 2)]
        for i in range(1, self.n_joints + 1):
            # i=1,...,6
            R_i_im1[i] = sp.Matrix([
                [c(i),                    -s(i),                   0],
                [s(i) * sp.cos(alpha[i-1]),  c(i) * sp.cos(alpha[i-1]),  -sp.sin(alpha[i-1])],
                [s(i) * sp.sin(alpha[i-1]),  c(i) * sp.sin(alpha[i-1]),  sp.cos(alpha[i-1])]
            ])

        # R_i_im1[1] means the orientation of frame 1 wrt to 0.
        # R_i_im1[4] means the orientation of frame 4 wrt to 3.
        # R_i_im1[4+1] = orientation of frame 5 to frame 4

        R_im1_i = [r.transpose() for r in R_i_im1]
        # R_im1_i[4] means the rotation from frame 3 to 4.

        return R_i_im1, R_im1_i

    def __rigid_body_dynamics(self):
        """
        Follows algorithm described in:
        Craig, John J. 2009. "Introduction to Robotics: Mechanics and Control", 3/E. Pearson Education India.
        """

        identity_3 = sp.eye(3)
        (m, d, a, alpha) = self.__mdh_num_to_sym()
        P = self.__get_P(a, d, alpha)  # position vectors
        PC = self.__pc  # center-of-mass locations

        I_CoM = [sp.zeros(3, 3) for _ in range(self.n_joints + 1)]
        for j in range(1, self.n_joints + 1):
            PC_dot_left = spdot(PC[j].transpose(), PC[j])
            PC_dot_right = spdot(PC[j], (PC[j].transpose()))
            assert PC_dot_left.shape == (1, 1)
            assert PC_dot_right.shape == (3, 3)
            PC_dot_scalar = PC_dot_left[0, 0]
            I_CoM[j] = self.__i_cor[j] - m[j] * (PC_dot_scalar * identity_3 - PC_dot_right)

        # State
        w = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]  # Angular velocity
        wd = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]  # Angular acceleration
        vd = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]  # Translational acceleration

        # Gravity
        vd[0] = -self.__g

        vcd = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]
        F = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]
        N = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]

        f = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]
        f.append(self.__f_tcp)

        n = [spvector([0, 0, 0]) for _ in range(self.n_joints + 1)]
        n.append(self.__n_tcp)

        Z = spvector([0, 0, 1])
        (R_i_im1, R_im1_i) = self.__get_forward_kinematics(alpha)

        # Outputs
        tau = [sp.zeros(1, 1) for _ in range(self.n_joints + 1)]

        # Outward recursion j: 0 -> 5
        for j in range(self.n_joints):
            w[j+1] = spdot(R_im1_i[j+1], w[j]) + self.qd[j+1] * Z
            assert w[j+1].shape == (3, 1)
            wd[j+1] = spdot(R_im1_i[j+1], wd[j]) + spcross(spdot(R_im1_i[j+1], w[j]), self.qd[j+1] * Z) + self.qdd[j+1] * Z
            assert wd[j+1].shape == (3, 1)
            vd[j+1] = spdot(R_im1_i[j+1], spcross(wd[j], P[j+1]) + spcross(w[j], spcross(w[j], P[j+1])) + vd[j])
            assert vd[j+1].shape == (3, 1)
            vcd[j+1] = spcross(wd[j+1], PC[j+1]) + spcross(w[j+1], spcross(w[j+1], PC[j+1])) + vd[j+1]
            assert vcd[j+1].shape == (3, 1)
            F[j+1] = m[j+1] * vcd[j+1]
            assert F[j+1].shape == (3, 1)
            N[j+1] = spdot(I_CoM[j+1], wd[j+1]) + spcross(w[j+1], spdot(I_CoM[j+1], w[j+1]))
            assert N[j+1].shape == (3, 1)

        # Inward recursion j: 6 -> 1
        for j in reversed(range(1, self.n_joints + 1)):
            f[j] = spdot(R_i_im1[j+1], f[j+1]) + F[j]
            n[j] = N[j] + spdot(R_i_im1[j+1], n[j+1]) + spcross(PC[j], F[j]) + spcross(P[j+1], spdot(R_i_im1[j+1], f[j+1]))
            assert n[j].shape == (3, 1), n[j]
            tau[j] = spdot(n[j].transpose(), Z)
            assert tau[j].shape == (1, 1)

        tau_list = [t[0, 0] for t in tau]

        return tau_list


    def plot_kinematics(self, block=True):
        try:
            import roboticstoolbox as rtb
            MDHRobot = self._create_kinematics_mdh(rtb)
            robot_kinematics = MDHRobot(self.mdh)
            robot_kinematics.plot(robot_kinematics.q, block=block)
        except ImportError:
            import warnings
            warnings.warn("The roboticstoolbox package is not installed, please install it for plotting the robot kinematics.")

    
    def _create_kinematics_mdh(self, rtb):
        class MDHRobot(rtb.DHRobot):
            def __init__(self, mdh: ModifiedDH):
                links = []
                n_links = []
                for i in range(mdh.n_joints):
                    d = mdh.d[i]
                    a = mdh.a[i]
                    alpha = mdh.alpha[i]
                    setattr(self, f"link_{i}", rtb.RevoluteMDH(d=d, a=a, alpha=alpha))
                    links.append(getattr(self, f"link_{i}"))
                    n_links.append(i+1)
                super().__init__(links, name="robot")
        return MDHRobot





import pickle

from aurt.rigid_body_dynamics import RigidBodyDynamics
from aurt.robot_dynamics import RobotDynamics
from aurt.robot_calibration import RobotCalibration
from aurt.robot_data import RobotData
from aurt.data_processing import *
from aurt.file_system import from_cache


def compile_rbd(mdh_path, gravity, output_path, plotting, block=True):
    mdh = convert_file_to_mdh(mdh_path)
    rbd = RigidBodyDynamics(mdh, gravity)
    rbd.regressor()

    if plotting:
        rbd.plot_kinematics(block)

    # save class
    pathfile = from_cache(output_path + ".pickle")
    with open(pathfile, 'wb') as f:
        pickle.dump(rbd, f)

def compile_rd(rbd_filename, friction_load_model, friction_viscous_powers, output_path):
    rd = RobotDynamics(rbd_filename, viscous_friction_powers=friction_viscous_powers, friction_load_model=friction_load_model)
    rd.regressor()

    # save class
    pathfile = from_cache(output_path + ".pickle")
    with open(pathfile, 'wb') as f:
        pickle.dump(rd, f)


def calibrate(model_path, data_path, output_params, output_calibration, plotting):
    rc = RobotCalibration(model_path, data_path)
    params = rc.calibrate(output_params)
    if plotting:
        rc.plot_calibration(params)

    # save class
    pathfile = from_cache(output_calibration + ".pickle")
    with open(pathfile, 'wb') as f:
        pickle.dump(rc, f)


def predict(model_path, data_path, output_path):
    filename = from_cache(model_path + ".pickle")
    with open(filename, 'rb') as f:
        rc: RobotCalibration = pickle.load(f)

    rc_predict_data = RobotData(data_path, delimiter=' ', interpolate_missing_samples=True) # TODO should we always interpolate missing samples?
    rc.predict(rc_predict_data, rc.parameters, output_path)



def calibrate_validate(model_path, data_path, calibration_data_relative, output_params, output_calibration, output_predict, plotting):
    rc = RobotCalibration(model_path, data_path, calibration_data_relative)
    params = rc.calibrate(output_params)
    output_pred = rc.predict(rc.robot_data_validation, params, output_predict)
    if plotting:
        rc.plot_calibrate_and_validate(params)

    # save class
    pathfile = from_cache(output_calibration + ".pickle")
    with open(pathfile, 'wb') as f:
        pickle.dump(rc, f)


import numpy as np


def find_first(x):
    """
    Finds first true value in the bool array x and returns the index of that value.
    """
    idx = x.view(bool).argmax() // x.itemsize
    return idx if x[idx] else -1


def find_nonstatic_start_and_end_indices(qd_target, qd_threshold=0.0001):
    """
    qd_target: (Njoints x n_samples) numpy array
    Finds start and end indices for useful (non-zero angular velocity) subset of data.
    Procedure:
    1.  Find first non-zero qd_target (starting from beginning and searching forward)
    2.  Find first non-zero qd_target (starting from end and searching backwards)
    Below is shown the angular velocity profile and the start and end indices.
                                       __
                      ^       ,___    /  \____
    qd_target [rad/s] |      /    \__/        \
                      |     /                  \
                      |____/____________________\____> sample no. [-]
                           ^idx_start           ^idx_end
    """

    idx_start = find_first(np.any(-qd_threshold > qd_target, axis=0) | np.any(qd_target > qd_threshold, axis=0))
    idx_start = max(1, idx_start)  # idx start has minimum value of 1 for estimation of double derivatives to work
    idx_end = len(qd_target[0, :]) - 1 - find_first(np.any(-qd_threshold > qd_target[:, ::-1], axis=0) | np.any(qd_target[:, ::-1] > qd_threshold, axis=0))
    assert 0 < idx_start < idx_end

    return idx_start, idx_end


def find_const_vel_start_and_end_indices(t, qd_target, qdd_threshold=0.01, t_stabilize=0.4):
    # Finds start and end indices for useful (constant angular velocity) subset of data.
    # TODO: If not qdd_target_new(idx_mid) == 0, analyse entire qdd_target_new and find the longest duration of
    #  qdd_target_new == 0 while (in addition) qd_target != 0.
    # Procedure:
    # 1.  Find center of time series data
    # 1a. Check if qdd(signal center) == 0 (constant angular velocity)
    #     From experience, the qdd_target provided by the UR controller is erroneous, thus we use:
    #          qdd_target = diff(qd_target)
    # 2.  Starting from the center of qdd_target, find the first nonzero qdd_target in each direction.
    # 3.  For the identified starting index, add a positive integer due to possible non-steady state/oscillations, which
    #     may violate the assumption of constant velocity.
    #
    # Below is shown the angular velocity profile and the start and end indices.
    #
    #                       oscillation(s)
    #                            |
    #                            v
    #                   ^       ,-_______________
    # qd_target [rad/s] |      /  :             :\
    #                   |     /   :             : \
    #                   |____/____:_____________:__\____> sample no. [-]
    #                             ^idx_start    ^idx_end
    idx_mid = t.size // 2
    dt = t[1] - t[0]
    n_stabilize = round(t_stabilize / dt)
    qdd_target = np.diff(qd_target, prepend=0.0) / dt
    assert (abs(qdd_target[idx_mid]) < qdd_threshold)
    idx_start = n_stabilize + idx_mid - find_first((-qdd_threshold > qdd_target[0:idx_mid][::-1]) | (
            qdd_target[0:idx_mid][::-1] > qdd_threshold))
    idx_end = idx_mid - 2 + find_first(
        (-qdd_threshold > qdd_target[idx_mid:-1]) | (qdd_target[idx_mid:-1] > qdd_threshold))

    assert idx_start < idx_end

    return idx_start, idx_end


def negate_negative(lst):  # return negated value of negative elements
    return [-x for x in lst if x < 0.0] or None


import sympy as sp
import numpy as np
import pickle

from aurt.file_system import cache_object, from_cache
from aurt.rigid_body_dynamics import RigidBodyDynamics
from aurt.joint_dynamics import JointDynamics


class RobotDynamics:
    def __init__(self, rbd_filename, tcp_force_torque=None, viscous_friction_powers=None, friction_load_model=None, hysteresis_model=None): # TODO ask Emil about tcp_force_torque
        # Load saved RigidBodyDynamics model
        filename = from_cache(rbd_filename + ".pickle")
        with open(filename, 'rb') as f:
            self.rigid_body_dynamics: RigidBodyDynamics = pickle.load(f)
        self.n_joints = self.rigid_body_dynamics.mdh.n_joints
        self.q = [sp.Integer(0)] + [sp.symbols(f"q{j}") for j in range(1, self.n_joints + 1)]
        self.qd = [sp.Integer(0)] + [sp.symbols(f"qd{j}") for j in range(1, self.n_joints + 1)]
        self.qdd = [sp.Integer(0)] + [sp.symbols(f"qdd{j}") for j in range(1, self.n_joints + 1)]
        self.tauJ = sp.symbols([f"tauJ{j}" for j in range(self.n_joints + 1)])

        self.joint_dynamics = JointDynamics(self.n_joints,
                                            load_model=friction_load_model,
                                            hysteresis_model=hysteresis_model,
                                            viscous_powers=viscous_friction_powers)

    def filepath_regressor_joint(self, j):
        return from_cache(f'robot_dynamics_regressor_joint_{j}')

    def parameters(self):
        """
        Returns a list of 'n_joints + 1' elements with each element comprising a list of all parameters related to that
        corresponding link and joint.
        """
        par_rigid_body = self.rigid_body_dynamics.params
        par_joint = self.joint_dynamics.parameters()
        return [par_rigid_body[j] + par_joint[j] for j in range(self.n_joints)]

    def number_of_parameters(self):
        n_par_rbd = self.rigid_body_dynamics.n_params
        n_par_jd = self.joint_dynamics.number_of_parameters()
        return [n_par_rbd[j] + n_par_jd[j] for j in range(self.n_joints)]

    def observation_matrix_joint(self, j, q_num, qd_num, qdd_num):
        assert q_num.shape == qd_num.shape == qdd_num.shape
        assert 0 <= j < self.n_joints

        n_samples = q_num.shape[1]
        obs_mat_j = np.empty((n_samples, 0))
        obs_mat_j_rbd = self.rigid_body_dynamics.observation_matrix_joint(j, q_num, qd_num, qdd_num)
        for j_par in range(self.n_joints):
            if j_par == j:
                tauJ_j_basis_num = np.sum(obs_mat_j_rbd, axis=1).transpose()
                obs_mat_j_jd = self.joint_dynamics.observation_matrix_joint(j, qd_num[j, :], tauJ_j_basis_num)
            else:
                obs_mat_j_jd = np.zeros((n_samples, self.joint_dynamics.number_of_parameters()[j_par]))

            col_start = sum(self.rigid_body_dynamics.n_params[:j_par])
            col_end = col_start + self.rigid_body_dynamics.n_params[j_par]
            obs_mat_j_jpar_rbd = obs_mat_j_rbd[:, col_start:col_end]
            obs_mat_j = np.hstack((obs_mat_j, obs_mat_j_jpar_rbd, obs_mat_j_jd))
        return obs_mat_j

    def observation_matrix(self, q_num, qd_num, qdd_num):
        assert q_num.shape == qd_num.shape == qdd_num.shape

        n_samples = q_num.shape[1]

        # Merge observation matrices for rigid-body dynamics and joint dynamcis
        observation_matrix = np.empty((self.n_joints*n_samples, sum(self.number_of_parameters())))
        for j in range(self.n_joints):
            observation_matrix[j*n_samples:(j+1)*n_samples, :] = self.observation_matrix_joint(j, q_num, qd_num, qdd_num)

        return observation_matrix

    def regressor_joint(self, j):
        """Constructs row 'j' of the regressor matrix."""
        def compute_regressor_joint():
            reg_j = sp.Matrix(1, 0, [])

            for j_par in range(self.n_joints):
                reg_j_rbd_par_j = self.rigid_body_dynamics.regressor_joint_parameters_for_joint(j, j_par)
                if j_par == j:
                    reg_j_jd = self.joint_dynamics.regressor()[j_par]
                else:
                    reg_j_jd = sp.zeros(1, self.joint_dynamics.number_of_parameters()[j_par])

                reg_j = sp.Matrix.hstack(reg_j, reg_j_rbd_par_j, reg_j_jd)
            return reg_j

        return compute_regressor_joint()#cache_object(self.filepath_regressor_joint(j+1), compute_regressor_joint)

    def regressor(self, output_filename="robot_dynamics_regressor"):
        filepath_regressor = from_cache(output_filename)

        def compute_regressor():
            """Merges regressor matrices for rigid-body dynamics and joint dynamics to construct a regressor for the
            robot dynamics."""
            reg = sp.zeros(self.n_joints, sum(self.number_of_parameters()))
            for j in range(self.n_joints):
                reg[j, :] = self.regressor_joint(j)

            # for j in range(self.n_joints):
            #     cache_object(from_cache(f"robot_dynamics_regressor_joint_{j+1}"), lambda: reg[j, :])
            return reg

        return cache_object(filepath_regressor, compute_regressor)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aurt",
    version="0.0.1",
    author="Emil Madsen, Daniella Tola, Claudio Gomes",
    author_email="ema@ece.au.dk",
    description="A robot dynamic parameters calibration toolbox.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.au.dk/software-engineering/aurt",
    packages=["aurt"],
    install_requires=[
        "numpy>=1",
        "sympy>=1",
        "pandas>=1",
        "scikit-learn>=0.24"
    ],
    extras_require={
        "vis": ["roboticstoolbox-python", "matplotlib>=1"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    entry_points={"console_scripts": ["aurt=aurt.cli:main"]},
)
